import numpy as np
from Node import BBNode


# Determine whether the relaxation solution is integral and the branching index
def relaxation_branching(z, lower_x, tol):
    if np.size(z) != 0:
        int_sol = (z + 0.5).astype(int)
        sol_diff = z - int_sol
        if np.max(abs(sol_diff)) > tol:
            loc = np.where(abs(sol_diff) > tol)[0]
            max_ind = np.argmax([lower_x[index] for index in loc])
            return loc[max_ind]
    return -1


def is_integral(z, lower_x, tol):
    return True if relaxation_branching(z, lower_x, tol) == -1 else False


# Update the sets S_0 and S_1 for children nodes
def new_z(node, index, delta):
    new_zlb = node.zlb + [index]
    new_zub = node.zub + [index]
    if index - delta < 0:
        new_zub_separation = node.zub + [x for x in range(0, index)] \
                             + [x for x in range(index + 1, index + delta + 1)] \
                             + [x for x in range(index - delta + node.a.shape[1], node.a.shape[1])]
    elif index + delta + 1 > node.a.shape[1]:
        new_zub_separation = node.zub + [x for x in range(index - delta, index)] \
                             + [x for x in range(index + 1, node.a.shape[1])] \
                             + [x for x in range(0, index + delta + 1 - node.a.shape[1])]
    else:
        new_zub_separation = node.zub + [x for x in range(index - delta, index)] \
                             + [x for x in range(index + 1, index + delta + 1)]
    return new_zlb, new_zub, new_zub_separation


# Create children nodes according to the updated sets S_0 and S_1
def branch(current_node, tol, delta):
    branching_variable = relaxation_branching(current_node.z, current_node.lower_x, tol)
    branching_variable = current_node.support[branching_variable]
    new_zlb, new_zub, new_zub_separation = new_z(current_node, branching_variable, delta)
    left_node = BBNode(current_node, new_zlb, new_zub_separation)
    right_node = BBNode(current_node, current_node.zlb, new_zub)
    return left_node, right_node
