import numpy as np
from gurobipy import Model, GRB, QuadExpr, LinExpr


# Use commercial solver gurobi to solve R
def l0gurobi(a, y, k, delta, u, lb, ub):
    # Set the optimization model
    model = Model()
    m = a.shape[0]
    n = a.shape[1]

    x = {}
    z = {}

    for column_index in range(n):
        x[column_index] = model.addVar(vtype=GRB.CONTINUOUS, name='x'+str(column_index), ub=u, lb=0)
        z[column_index] = model.addVar(vtype=GRB.CONTINUOUS, name='z'+str(column_index),
                                       ub=ub[column_index], lb=lb[column_index])

    r = {}
    for row_index in range(m):
        r[row_index] = model.addVar(vtype=GRB.CONTINUOUS, name='r'+str(row_index), ub=GRB.INFINITY, lb=-GRB.INFINITY)

    model.update()

    # Least Square Objective
    obj = QuadExpr()

    for row_index in range(m):
        obj.addTerms(0.5, r[row_index], r[row_index])

    model.setObjective(obj, GRB.MINIMIZE)

    # Constraints
    for row_index in range(m):
        expr1 = LinExpr()
        expr1.addTerms(a[row_index, :], [x[column_index] for column_index in range(n)])
        model.addConstr(r[row_index] == y[row_index] - expr1)

    for column_index in range(n):
        model.addConstr(x[column_index] <= z[column_index] * u)
        model.addConstr(x[column_index] >= 0)

    expr2 = LinExpr()
    for column_index in range(n):
        expr2.addTerms(1.0, z[column_index])
    model.addConstr(expr2 <= k)

    for column_index in range(n-delta):
        expr3 = LinExpr()
        expr3.addTerms([1.0 for _ in range(delta + 1)],
                       [z[index] for index in range(column_index, column_index + delta + 1)])
        model.addConstr(expr3 <= 1)

    for column_index in range(n-delta, n):
        expr4 = LinExpr()
        expr4.addTerms([1.0 for _ in range(column_index - n + delta + 1)],
                       [z[index] for index in range(column_index - n + delta + 1)])
        expr4.addTerms([1.0 for _ in range(n-column_index)],
                       [z[index] for index in range(column_index, n)])
        model.addConstr(expr4 <= 1)

    model.update()
    model.optimize()

    output_x = np.zeros(len(x))
    output_z = np.zeros(len(z))

    for i in range(len(x)):
        output_x[i] = x[i].x
        output_z[i] = z[i].x

    return output_x, output_z, model.ObjVal
