import numpy as np
from Gurobi import l0gurobi
from Perspective_Formulation import perspective
from Relaxation import relaxation


class BBNode:
    def __init__(self, parent, zlb, zub, **kwargs):
        """
        Construct a node in BBTree

        Inputs:
        parent: the parent node
        zlb: the lower bound of the auxiliary integer variables z
        zub: the upper bound of the auxiliary integer variables z

        kwargs: dictionary of other arguments (used for root node once)
        a: measurement matrix
        y: measurements
        """

        self.a = kwargs.get('a', parent.a if parent else None)
        self.y = kwargs.get('y', parent.y if parent else None)
        self.parent_lower_bound = parent.lower_bound if parent else None

        self.level = parent.level + 1 if parent else 0

        self.zlb = zlb
        self.zub = zub
        self.z = None

        self.upper_bound = None
        self.lower_bound = None

        self.support = None
        self.upper_x = None
        self.lower_x = None

    # Solve the relaxations by open source packages or commercial solvers gurobi
    def lower_solve(self, k, delta, u, lower_solver, int_tol=1e-6):
        full_zlb = np.zeros(self.a.shape[1])
        full_zlb[self.zlb] = 1
        full_zub = np.ones(self.a.shape[1])
        full_zub[self.zub] = 0
        if lower_solver == 'cvx':
            lower_x, z, self.lower_bound = relaxation(self.a, self.y, k, delta, u, full_zlb, full_zub)
        elif lower_solver == 'perspective':
            lower_x, z, self.lower_bound = perspective(self.a, self.y, k, delta, u, full_zlb, full_zub)
        elif lower_solver == 'gurobi':
            lower_x, z, self.lower_bound = l0gurobi(self.a, self.y, k, delta, u, full_zlb, full_zub)
        else:
            pass
        self.support = list(np.where(lower_x > int_tol)[0])
        self.upper_x = lower_x
        self.lower_x = lower_x[self.support]
        self.z = z[self.support]
        return self.lower_bound

