import numpy as np
from cvxopt import matrix, solvers


# Use open-source package cvxopt to solve R
def relaxation(a, y, k, delta, u, lb, ub):
    m = a.shape[0]
    n = a.shape[1]
    a_z = np.zeros((m, n))
    a_full = np.concatenate((a, a_z), axis=1)
    p = np.dot(np.transpose(a_full), a_full)
    q = -np.dot(np.transpose(y), a_full)

    # Generate inequality constraint matrix g (left-side)
    g = np.zeros((5*n+1, 2*n))
    g[0, 0:n] = 0
    g[0, n:2*n] = 1
    for i in range(0, n-delta):
        g[i+1, i+n:(i+n+delta+1)] = 1
    for i in range(n-delta, n):
        g[i+1, i+n:2*n] = 1
        g[i+1, n:(i+delta+1)] = 1
    for i in range(0, n):
        g[i+n+1, i] = 1
        g[i+n+1, i+n] = -u
    for i in range(0, 2*n):
        g[i+2*n+1, i] = -1
    for i in range(0, n):
        g[i+4*n+1, i+n] = 1

    # Generate inequality constraint vector h (right-side)
    h = np.zeros((5*n+1, ))
    h[0] = k
    h[1:n+1] = 1
    h[n+1:2*n+1] = 0
    h[2*n+1:3*n+1] = 0
    h[3*n+1:4*n+1] = -lb
    h[4*n+1:5*n+1] = ub

    p = p.astype(float)
    q = q.astype(float)
    g = g.astype(float)
    h = h.astype(float)

    p = matrix(p)
    q = matrix(q)
    g = matrix(g)
    h = matrix(h)

    # Solve R
    solvers.options['show_progress'] = False
    sol = solvers.qp(p, q, g, h)
    arr = np.array(sol['x']).flatten()
    lower_x = arr[0:n]
    z = arr[n:2*n]
    obj = np.array(sol['primal objective']).flatten()[0] + 1/2*np.dot(np.transpose(y), y)

    # Return results
    return lower_x, z, obj
